#include <Regexp.h>
#include <ArduinoJson.h>
#include <string>
#include <unordered_map>
#include <unordered_set>
#include <errno.h>
#include <RHReliableDatagram.h>
#include <RH_RF95.h>
#include <SPI.h>
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>

#define LORA_DELAY 4000

#define JSON_CAPACITY 2048
#define MAX_JSON_FIELD_LEN 32
#define REGISTRATION_BUF_SIZE 32
#define NUMBUCKETS 131

#define COMMAND 0
#define BYTE_COUNT 1
#define START_PAYLOAD 2

#define NOP 0
#define REGISTRATION 1
#define CLEAR_REGISTRY 2
#define TELEMETRY 3

#define INTEGER 0x00
#define CUSTOM_FLOAT 0x40 
#define STRING 0x80
#define SPECIAL 0xC0
#define _8BIT 0x00
#define _16BIT 0x10 
#define _32BIT 0x20 
#define IP_ADDR 0x01
#define MAC_ADDR 0x02
#define DATE_YYYYMMDD 0x03
#define TIME_HHMMSS 0x04
#define JSON_OBJECT 0x05
#define GENERAL_TYPE_MASK 0xC0
#define SPECIAL_TYPE_MASK 0x3F
#define NUMBER_SIZE_MASK 0x30 
#define DECIMAL_POINT_MASK 0x0F

#define JSON_LEVEL_UP 0xFF

#define LORA_ROUTER 1
#define LORA_GATEWAY 2

#define MAX_AP_CONNECTIONS 8

const char* ssid     = "***";
const char* password = "***";

ESP8266WebServer server(80);

IPAddress local_IP(188,35,161,31);
IPAddress gateway(188,35,161,1);
IPAddress subnet(255,255,255,0);

RH_RF95 driver(D10, D4); // соединить проводом пины D2 и D4

RHReliableDatagram manager(driver, LORA_ROUTER);

StaticJsonDocument<JSON_CAPACITY> doc;

std::unordered_map<std::string, uint8_t> registry(NUMBUCKETS);
char registration_buf[REGISTRATION_BUF_SIZE][MAX_JSON_FIELD_LEN];
uint8_t registration_buf_count = 0;
std::unordered_set<std::string> registration_buf_set(31);


uint8_t send_buf[RH_RF95_MAX_MESSAGE_LEN];
uint8_t recv_buf[RH_RF95_MAX_MESSAGE_LEN];

void handleJSON() 
{
  server.send(200, "text/plain", "hello from LoRa router!");
  Serial.println(server.arg("plain"));

  DeserializationError err = deserializeJson(doc, server.arg("plain"));
  if (err) 
  {
    Serial.print(F("deserializeJson() failed with code "));
    Serial.println(err.f_str());
    return;
  }
  JsonObject obj = doc.as<JsonObject>();
  
  process_json(obj);  
}



void process_json(const JsonObject &obj)
{
  for (JsonPair kv : obj) {
    
    const char *key = kv.key().c_str();
    /*Serial.print(key);
    Serial.print(" : ");
    Serial.println(kv.value().as<String>());*/
    if(registry.count(std::string(key)) == 0)
    {
      if(send_buf[COMMAND] != REGISTRATION)
      {
        send_buf[COMMAND] = REGISTRATION;
        send_buf[BYTE_COUNT] = START_PAYLOAD;
        registration_buf_count = 0;
        registration_buf_set.clear();
        Serial.println(F("Requesting registration"));
      }
      
      if(registration_buf_set.count(std::string(key)) == 0)
      {
        uint8_t key_len = strlen(key) + 1;
        if(key_len <= MAX_JSON_FIELD_LEN && send_buf[BYTE_COUNT] + key_len <= sizeof(send_buf)
           && registration_buf_count < REGISTRATION_BUF_SIZE)
        {
          registration_buf_set.insert(std::string(key));
          
          /*Serial.print(key);
          Serial.print(" len = ");
          Serial.println(key_len);*/
          strcpy((char*)&send_buf[send_buf[BYTE_COUNT]], key); 
          send_buf[BYTE_COUNT] += key_len;
          strcpy(registration_buf[registration_buf_count++], key);
        }
        else
          return;
      }
         
    }
    else if (send_buf[COMMAND] != REGISTRATION &&
             send_buf[COMMAND] != TELEMETRY)
    {
      send_buf[COMMAND] = TELEMETRY;
      send_buf[BYTE_COUNT] = START_PAYLOAD;
    }
                      
    if (kv.value().is<JsonObject>())
    {
      if(send_buf[COMMAND] == TELEMETRY)
      {
        if (send_buf[BYTE_COUNT]+2 <= sizeof(send_buf))
        {
          send_buf[send_buf[BYTE_COUNT]++] = registry[std::string(key)];
          uint8_t tag = 0x00;
          tag |= SPECIAL | JSON_OBJECT;
          send_buf[send_buf[BYTE_COUNT]++] = tag;
        }
        else
          return;
      }
      process_json(kv.value());
      
      if(send_buf[COMMAND] == TELEMETRY)
      {
        if(send_buf[BYTE_COUNT]+1 <= sizeof(send_buf))
          send_buf[send_buf[BYTE_COUNT]++] = JSON_LEVEL_UP;  
        else
          return;
      }
           
    }
    else if(send_buf[COMMAND] == TELEMETRY)
    {
      String value = kv.value().as<String>();
      analyze_json_value(key, value.c_str());
    }
      
  }
}

void setup() {
  Serial.begin(9600);
  while (!Serial);
  
  if (!manager.init())
    Serial.println(F("init failed"));
  manager.setTimeout(LORA_DELAY);
  
  driver.setFrequency(868.9);
  driver.setTxPower(13, false);


  Serial.println();

  registry.clear();

  Serial.print(F("Configuring access point..."));
  WiFi.softAP(ssid, password, 1, false, MAX_AP_CONNECTIONS);
  WiFi.softAPConfig(local_IP, gateway, subnet);
  //WiFi.setOutputPower(15);
 
  IPAddress IP = WiFi.softAPIP();
  Serial.print(F("Access point IP address: "));
  Serial.println(IP);

  server.on("/core/jsonadd.php", handleJSON);

  server.begin();
  Serial.println(F("server started"));

}

void print_registry()
{
  Serial.println(F("Registry: "));
  for(auto iter = registry.begin(); iter != registry.end(); ++iter)
  {
    Serial.print(iter->first.c_str());
    Serial.print(": ");
    Serial.println((int)iter->second);
  }
  
}

void loop() {

  send_buf[COMMAND] = NOP;
  send_buf[BYTE_COUNT] = START_PAYLOAD;
  
  server.handleClient();

  if (send_buf[COMMAND] == NOP)
    return;

  Serial.print(F("Bytes to be sent : "));
  Serial.println(send_buf[BYTE_COUNT]);
 
  if (manager.sendtoWait(send_buf, send_buf[BYTE_COUNT], LORA_GATEWAY))
  {
   
    uint8_t len = sizeof(recv_buf);
    uint8_t from;   
     if (manager.recvfromAckTimeout(recv_buf, &len, LORA_DELAY * 3, &from) && from == LORA_GATEWAY)
     {
       Serial.println(F("Got reply from LoRa gateway"));
       switch (recv_buf[COMMAND])
       {
         case REGISTRATION:                
           for(int i = 0; i < recv_buf[BYTE_COUNT] - START_PAYLOAD; ++i)
           {
             registry.insert({std::string(registration_buf[i]), recv_buf[START_PAYLOAD+i]});
           }
           Serial.println(F("Registration complete"));
           print_registry();
           break;   
         case CLEAR_REGISTRY:   
           Serial.println(F("Got registry clear request"));      
           registry.clear();
           Serial.println(F("Registry cleared"));
           break;     
        }
     }
     else
     {
       Serial.println(F("No reply, is LoRa gateway running?"));
     }    
  }
  else
    Serial.println(F("sendtoWait failed")); 
  
  delay(LORA_DELAY);  
  
}

void analyze_json_value(const char *key, const char *v)
{
  if(v == NULL)
    return;
  uint8_t key_id = registry[std::string(key)];
  Serial.print(key);
  Serial.print(F(" : "));
  Serial.print(v);
  Serial.print(F(" is a "));
  char value[MAX_JSON_FIELD_LEN];
  strlcpy(value, v, MAX_JSON_FIELD_LEN);
  uint8_t tag = 0x00;
  
  MatchState ms;
  ms.Target(value);  
  if(ms.Match("^%s*%-?%d+%.%d+%s*$") == REGEXP_MATCHED)
  {
    Serial.print(F("custom float "));
    tag |= CUSTOM_FLOAT;

    char integ[13], frac[16];
    int ret = sscanf(value, " %12[0-9-].%15[0-9]", &integ, &frac);
    if (ret != 2)
    {
      Serial.println(F("; input error, sending as string"));
      goto send_as_str;      
    }
    uint8_t decimal_places = strlen(frac);
    tag |= decimal_places;
    
    char integfrac[28];
    strcpy(integfrac, integ);
    strcat(integfrac, frac);      

    errno = 0;
    long equiv_int = strtol(integfrac, NULL, 10);
    if (errno != 0)
    {
      Serial.println(F("; conversion error, sending as string"));
      goto send_as_str;
    }
    
    if (INT8_MIN <= equiv_int && equiv_int <= INT8_MAX)
    {
      Serial.println(F("8 bit"));
      tag |= _8BIT;
      if(send_buf[BYTE_COUNT]+2+1 <= sizeof(send_buf))
      {
        send_buf[send_buf[BYTE_COUNT]++] = key_id;
        send_buf[send_buf[BYTE_COUNT]++] = tag;
        
        send_buf[send_buf[BYTE_COUNT]++] = equiv_int;
        
      }
    }
    else if (INT16_MIN <= equiv_int && equiv_int <= INT16_MAX)
    {
      Serial.println(F("16 bit"));
      tag |= _16BIT;
      if(send_buf[BYTE_COUNT]+2+2 <= sizeof(send_buf))
      {
        send_buf[send_buf[BYTE_COUNT]++] = key_id;
        send_buf[send_buf[BYTE_COUNT]++] = tag;
        send_buf[send_buf[BYTE_COUNT]++] = equiv_int & 0x00FF;
        send_buf[send_buf[BYTE_COUNT]++] = (equiv_int & 0xFF00) >> 8;
      }
    }
    else if (INT32_MIN <= equiv_int && equiv_int <= INT32_MAX)
    {
      Serial.println(F("32 bit"));
      tag |= _32BIT;
      if(send_buf[BYTE_COUNT]+2+4 <= sizeof(send_buf))
      {
        send_buf[send_buf[BYTE_COUNT]++] = key_id;
        send_buf[send_buf[BYTE_COUNT]++] = tag;
        send_buf[send_buf[BYTE_COUNT]++] = equiv_int & 0x000000FF;
        send_buf[send_buf[BYTE_COUNT]++] = (equiv_int & 0x0000FF00) >> 8;
        send_buf[send_buf[BYTE_COUNT]++] = (equiv_int & 0x00FF0000) >> 16;
        send_buf[send_buf[BYTE_COUNT]++] = (equiv_int & 0xFF000000) >> 24;
      }
    }
    
  }
  else if (ms.Match("^%s*%-?%d+%s*$") == REGEXP_MATCHED)
  {
    Serial.print(F("integer "));
    tag |= INTEGER;

    errno = 0;
    long num = strtol(value, NULL, 10);
    if (errno != 0)
    {
      Serial.println(F("; conversion error, sending as string"));
      goto send_as_str;
    }
    
    if (INT8_MIN <= num && num <= INT8_MAX)
    {
      Serial.println(F("8 bit"));
      tag |= _8BIT;
      if(send_buf[BYTE_COUNT]+2+1 <= sizeof(send_buf))
      {
        send_buf[send_buf[BYTE_COUNT]++] = key_id;
        send_buf[send_buf[BYTE_COUNT]++] = tag;
        
        send_buf[send_buf[BYTE_COUNT]++] = num;
        
      }
    }
    else if (INT16_MIN <= num && num <= INT16_MAX)
    {
      Serial.println(F("16 bit"));
      tag |= _16BIT;
      if(send_buf[BYTE_COUNT]+2+2 <= sizeof(send_buf))
      {
        send_buf[send_buf[BYTE_COUNT]++] = key_id;
        send_buf[send_buf[BYTE_COUNT]++] = tag;
        send_buf[send_buf[BYTE_COUNT]++] = num & 0x00FF;
        send_buf[send_buf[BYTE_COUNT]++] = (num & 0xFF00) >> 8;
      }
    }
    else if (INT32_MIN <= num && num <= INT32_MAX)
    {
      Serial.println(F("32 bit"));
      tag |= _32BIT;
      if(send_buf[BYTE_COUNT]+2+4 <= sizeof(send_buf))
      {
        send_buf[send_buf[BYTE_COUNT]++] = key_id;
        send_buf[send_buf[BYTE_COUNT]++] = tag;
        send_buf[send_buf[BYTE_COUNT]++] = num & 0x000000FF;
        send_buf[send_buf[BYTE_COUNT]++] = (num & 0x0000FF00) >> 8;
        send_buf[send_buf[BYTE_COUNT]++] = (num & 0x00FF0000) >> 16;
        send_buf[send_buf[BYTE_COUNT]++] = (num & 0xFF000000) >> 24;
        
      }
    }
   
  }
  else if (ms.Match("^%s*%x%x:%x%x:%x%x:%x%x:%x%x:%x%x%s*$") == REGEXP_MATCHED)
  {
    Serial.println(F("MAC"));
    tag |= SPECIAL | MAC_ADDR;
    unsigned char mac[6];
    sscanf(value, " %hhx:%hhx:%hhx:%hhx:%hhx:%hhx", &mac[0], &mac[1], &mac[2], &mac[3], &mac[4], &mac[5]);
    if(send_buf[BYTE_COUNT]+2+6 <= sizeof(send_buf))
    {
      send_buf[send_buf[BYTE_COUNT]++] = key_id;
      send_buf[send_buf[BYTE_COUNT]++] = tag;
      send_buf[send_buf[BYTE_COUNT]++] = mac[0];
      send_buf[send_buf[BYTE_COUNT]++] = mac[1];
      send_buf[send_buf[BYTE_COUNT]++] = mac[2];
      send_buf[send_buf[BYTE_COUNT]++] = mac[3];
      send_buf[send_buf[BYTE_COUNT]++] = mac[4];
      send_buf[send_buf[BYTE_COUNT]++] = mac[5];
    }   
  }
  else if(ms.Match("^%s*%d?%d?%d%.%d?%d?%d%.%d?%d?%d%.%d?%d?%d%s*$") == REGEXP_MATCHED)
  {
    Serial.println(F("IP"));
    tag |= SPECIAL | IP_ADDR;
    unsigned char ip[4];
    sscanf(value, " %hhu.%hhu.%hhu.%hhu", &ip[0], &ip[1], &ip[2], &ip[3]);
    
    if(send_buf[BYTE_COUNT]+2+4 <= sizeof(send_buf))
    {
      send_buf[send_buf[BYTE_COUNT]++] = key_id;
      send_buf[send_buf[BYTE_COUNT]++] = tag;
      send_buf[send_buf[BYTE_COUNT]++] = ip[0];
      send_buf[send_buf[BYTE_COUNT]++] = ip[1];
      send_buf[send_buf[BYTE_COUNT]++] = ip[2];
      send_buf[send_buf[BYTE_COUNT]++] = ip[3];
    }   
  }
  else if(ms.Match("^%s*%d%d%d%d%-%d%d%-%d%d%s*$") == REGEXP_MATCHED)
  {
    Serial.println(F("date"));
    tag |= SPECIAL | DATE_YYYYMMDD;
    unsigned int year;
    unsigned char month, day;
    
    sscanf(value, " %u-%hhu-%hhu", &year, &month, &day);
    if(send_buf[BYTE_COUNT]+2+2+1+1 <= sizeof(send_buf))
    {
      send_buf[send_buf[BYTE_COUNT]++] = key_id;
      send_buf[send_buf[BYTE_COUNT]++] = tag;
      send_buf[send_buf[BYTE_COUNT]++] = year & 0x00FF;     
      send_buf[send_buf[BYTE_COUNT]++] = (year & 0xFF00) >> 8;
      send_buf[send_buf[BYTE_COUNT]++] = month;
      send_buf[send_buf[BYTE_COUNT]++] = day;
    }
  }
  else if(ms.Match("^%s*%d%d:%d%d:%d%d%s*$") == REGEXP_MATCHED)
  {
    Serial.println(F("time"));
    tag |= SPECIAL | TIME_HHMMSS;
    unsigned char hours, minutes, seconds;
    sscanf(value, " %hhu:%hhu:%hhu", &hours, &minutes, &seconds);
    if(send_buf[BYTE_COUNT]+2+1+1+1 <= sizeof(send_buf))
    {
      send_buf[send_buf[BYTE_COUNT]++] = key_id;
      send_buf[send_buf[BYTE_COUNT]++] = tag;
      send_buf[send_buf[BYTE_COUNT]++] = hours;
      send_buf[send_buf[BYTE_COUNT]++] = minutes;
      send_buf[send_buf[BYTE_COUNT]++] = seconds;
      
    }
  }
  else
  {
send_as_str:
    Serial.println(F("string"));

    tag = STRING;
    uint8_t value_len = strlen(value) + 1;
    if (send_buf[BYTE_COUNT] + 2 + value_len <= sizeof(send_buf))
        {         
          send_buf[send_buf[BYTE_COUNT]++] = key_id;
          send_buf[send_buf[BYTE_COUNT]++] = tag;
          strcpy((char*)&send_buf[send_buf[BYTE_COUNT]], value); 
          send_buf[BYTE_COUNT] += value_len;        
        }
  } 
  
}
